//##################################################################
// DumpModules (builds in either x86 or x64 mode)
//##################################################################

#include DumpModules.h

DATA SECTION
	hInstance	HANDLE		?

CODE SECTION

START:
	invoke GetModuleHandle, 0
	mov [hInstance],rax
	invoke InitCommonControls
	INVOKE DialogBoxParam,[hInstance],1000,0,OFFSET DlgProc,0
	invoke ExitProcess,0

DlgProc FRAME hwnd,uMsg,wParam,lParam
	uses rbx
	LOCAL rc:RECT

	cmp D[uMsg],WM_INITDIALOG
	jne >>.WM_COMMAND
		invoke GetClientRect,[hwnd],offset rc
		invoke GetDlgItem,[hwnd],1001
		mov rbx,rax

		invoke SetWindowPos,rbx,NULL,[rc.left],[rc.top],[rc.right],[rc.bottom],SWP_NOZORDER

		invoke SetupDllListView,rbx
		invoke WalkLoadedModules,rbx

		// Sort the listview based on the import type
		invoke SendMessage,rbx,LVM_SORTITEMSEX,rbx,OFFSET ListViewSortProc

		jmp >.EXIT

	.WM_COMMAND
	cmp D[uMsg],WM_COMMAND
	jne >.WM_CLOSE

		jmp >.EXIT

	.WM_CLOSE
	cmp D[uMsg],WM_CLOSE
	jne >.DEFPROC
		INVOKE EndDialog,[hwnd],0

	.DEFPROC
		mov rax,FALSE
		ret

	.EXIT

	mov rax, TRUE
	ret
ENDF

MapPEFile FRAME pszFilename, phFile, phMapfile, ppMapfile
	uses rbx,rdi,rsi

	// Get a file handle
	invoke CreateFile,[pszFilename],GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,NULL,NULL
	mov rbx,[phFile]
	mov [rbx],rax
	test rax,rax
	jns >
		ret
	:

	// Map the file
	invoke CreateFileMapping,[rbx],NULL,PAGE_READONLY,0,0,NULL
	mov rdi,[phMapfile]
	mov [rdi],rax

	// Check the return value of CreateFileMapping to see if there was an error
	test rax,rax
	jnz >
		invoke CloseHandle,[rbx]
		xor rax,rax
		dec rax
		ret
	:

	// Get the address of the mapping
	invoke MapViewOfFile,[rdi],FILE_MAP_READ,0,0,0
	mov rsi,[ppMapfile]
	mov [rsi],rax

	// Check the return value of MapViewOfFile to see if there was an error
	test rax,rax
	jnz >
		invoke CloseHandle,[rdi]
		invoke CloseHandle,[rbx]
		xor rax,rax
		dec rax
		ret
	:

	xor rax,rax
	ret

ENDF

UmmapPEFile FRAME hFile, hMapfile, pMapfile
	// Unmap the file and close the handles
	invoke UnmapViewOfFile,[pMapfile]
	invoke CloseHandle,[hMapfile]
	invoke CloseHandle,[hFile]
	RET
ENDF

WalkLoadedModules FRAME hListview
	uses rbx,rdi,rsi
	LOCAL hProcess:%HANDLE
	LOCAL cbNeeded:D
	LOCAL pModuleArray:%PTR
	LOCAL BaseName[MAX_PATH]:%CHAR
	LOCAL hFile:%HANDLE
	LOCAL hMapfile:%HANDLE
	LOCAL pMapfile:%PTR
	LOCAL mi:MODULEINFO

	// Get this processes executable name and map the PE file
	invoke GetModuleFileName,NULL,offset BaseName,MAX_PATH
	invoke MapPEFile,offset BaseName,offset hFile,offset hMapfile,offset pMapfile
	// Be sure the file was mapped
	test eax,eax
	jns >
		// Skip everything if we can't map the file
		ret
	:

	// This is always -1 but just in case Microsoft ever changes it
	invoke GetCurrentProcess
	mov [hProcess],rax

	// The first call to the enumerator will return the amount of memory we need
	invoke EnumProcessModules,[hProcess],0,0,offset cbNeeded

	// Allocate the memory
	invoke VirtualAlloc,NULL,[cbNeeded],MEM_COMMIT,PAGE_READWRITE
	mov [pModuleArray],rax
	mov rdi,rax
	mov rbx,[cbNeeded]

	// Now we enumerate the modules for real
	invoke EnumProcessModules,[hProcess],rdi,rbx,offset cbNeeded
	:
		// Walk through the array of handles
		sub rbx,SIZEOF %HANDLE
		mov rax,[rdi+rbx]
		push rax
		// Get the name of the module
		invoke GetModuleBaseName,[hProcess],rax,offset BaseName,MAX_PATH
		pop rax
		// Get some information about the module
		invoke GetModuleInformation,[hProcess],rax,offset mi,SIZEOF MODULEINFO

		// See if this is a direct import
		// FindModuleInPE returns -1 if the module is not found, 0 if its found
		mov rsi,offset szExternal
		invoke FindModuleInPE,[pMapfile],offset BaseName
		test rax,rax
		js >.EXTRN
			mov rsi,offset szInternal
		.EXTRN

		// Add this module to the listview
		invoke AddToListview,[hListview],offset BaseName,[mi.lpBaseOfDll],rsi
		// Skip module #0, its this file
		cmp rbx,SIZEOF %HANDLE *2
		jge <<

	.EXIT
	// Free the array of handles
	invoke VirtualFree,[pModuleArray],NULL,MEM_RELEASE
	// Unmap the PE file
	invoke UmmapPEFile,[hFile],[hMapfile],[pMapfile]

	ret
	szExternal: CHAR A"External",0
	szInternal: CHAR A"Direct",0
endf

FindModuleInPE FRAME pMapFile, pFindFile
	uses rsi,rdi,rbx

	// Use the DbgHelp API functions for our walk through the PE
	// They do a good job and work in both 32 and 64 bit builds

	// Get the address of the IMAGE_NT_HEADERS structure
	invoke ImageNtHeader,[pMapFile]
	mov rsi,rax

	// Find the import directory's RVA
	#IF X64
		// For X64 builds use the 64 bit header
		mov rdi,[rsi+IMAGE_NT_HEADERS64.OptionalHeader.DataDirectory+8]
	#ELSE
		// For X32 builds use the 32 bit header
		mov rdi,[rsi+IMAGE_NT_HEADERS.OptionalHeader.DataDirectory+8]
	#ENDIF

	// Convert the RVA to a memory address in the mapped file
	invoke ImageRvaToVa,rsi,[pMapFile],rdi,NULL
	mov rdi,rax

	jmp >>L2
	L1:
		// Walk the PE import directory
		invoke ImageRvaToVa,rsi,[pMapFile],[rdi+IMAGE_IMPORT_DESCRIPTOR.Name],NULL
		mov rbx,rax
		// Is this the module we're looking for
		invoke lstrcmpi,[pFindFile],rbx
		test rax,rax
		jnz >
		// The module was found, return 0
		xor rax,rax
		ret
		:
		// Next descriptor
		add rdi,sizeof IMAGE_IMPORT_DESCRIPTOR

	L2:
	// If any of the descriptors is 0 then we're done
	cmp D[rdi+IMAGE_IMPORT_DESCRIPTOR.OriginalFirstThunk],0
	jne <<L1
	cmp D[rdi+IMAGE_IMPORT_DESCRIPTOR.TimeDateStamp],0
	jne <<L1
	cmp D[rdi+IMAGE_IMPORT_DESCRIPTOR.ForwarderChain],0
	jne <<L1
	cmp D[rdi+IMAGE_IMPORT_DESCRIPTOR.Name],0
	jne <<L1
	cmp D[rdi+IMAGE_IMPORT_DESCRIPTOR.FirstThunk],0
	jne <<L1

	// We've walked the whole directory and the module isn't there, return -1
	xor rax,rax
	dec rax
	ret
ENDF

AddToListview FRAME hListview,pszModule,LoadAddr,pszType
	LOCAL lvi:LVITEM
	LOCAL szLoadAddr[64]:%CHAR

	// Standard routine to add an item to the listview

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],0
	mov rax,[pszModule]
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_INSERTITEM,0,offset lvi

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],1
	invoke wsprintf,offset szLoadAddr,"%p",[LoadAddr]
	#IF !X64
		// For 32 bit builds we have to reset ESP
		add esp,12
	#ENDIF
	lea rax,szLoadAddr
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_SETITEM,0,offset lvi

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],2
	mov rax,[pszType]
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_SETITEM,0,offset lvi

	ret
endf

SetupDllListView FRAME hListview
	uses edi,esi
	LOCAL LVColumn:LVCOLUMN
	LOCAL hFont:%HANDLE

	// Use a fixed font, makes the addresses look nicer
	invoke GetStockObject,ANSI_FIXED_FONT
	mov [hFont],rax

	invoke SendMessage,[hListview],WM_SETFONT,[hFont],TRUE

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],200
	push ADDR "Module name"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,0,OFFSET LVColumn

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],150
	push ADDR "Load Address"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,1,OFFSET LVColumn

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],80
	push ADDR "Type"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,2,OFFSET LVColumn

	invoke SendMessage,[hListview],LVM_SETEXTENDEDLISTVIEWSTYLE,LVS_EX_GRIDLINES + LVS_EX_FULLROWSELECT,LVS_EX_FULLROWSELECT + LVS_EX_GRIDLINES

	ret
ENDF

ListViewSortProc FRAME lParam1,lParam2,lParamSort
	uses ebx,edi,esi
	LOCAL	ItemData1		:LVITEM
	LOCAL	ItemData2		:LVITEM
	LOCAL	Tbuff1[256]		:%CHAR
	LOCAL	Tbuff2[256]		:%CHAR

	mov rax,[lParam1]
	mov [ItemData1.iItem],rax
	mov D[ItemData1.iSubItem],2
	mov D[ItemData1.mask],LVIF_TEXT
	lea rax,Tbuff1
	mov [ItemData1.pszText],rax
	mov D[ItemData1.cchTextMax],255
	invoke SendMessage,[lParamSort],LVM_GETITEM,0,OFFSET ItemData1

	mov rax,[lParam2]
	mov [ItemData2.iItem],rax
	mov D[ItemData2.iSubItem],2
	mov D[ItemData2.mask],LVIF_TEXT
	lea rax,Tbuff2
	mov [ItemData2.pszText],rax
	mov D[ItemData2.cchTextMax],255
	invoke SendMessage,[lParamSort],LVM_GETITEM,0,OFFSET ItemData2

	invoke lstrcmpi,offset Tbuff1,offset Tbuff2
	RET
ENDF
